﻿// Apex4.js
//  This module, introduced in v0.21.x handles:
//      - Reading of the Apex 4 configuration file
//      - Instantiating the Apex 4 connection (and disconnection)
//      - Reporting on the status of the Apex 4 connection
//

// Dependencies and Globals
//#region Dependencies and Globals
var fs = require('fs');
var cp = require('child_process');
var exec = require('child_process').exec;

var Apex4HostInfo = null;// Holds the Address, Port and Token information for Apex 4
var Apex4Structure = { "token": "", "Inputs": [], "Outputs": [] };// Default I/O structure
var Apex4SocketConnection = null;// Socket.io-client Connection to Apex 4

// Apex 4 socket status
var Apex4SocketConnectionStatus = APEX4_SOCKET_UNKNOWN;// 
var APEX4_SOCKET_NOTDEFINED = 0;
var APEX4_SOCKET_UNKNOWN = 1;
var APEX4_SOCKET_CONNECTED = 2;
var APEX4_SOCKET_DISCONNECTED = 3;
var APEX4_SOCKET_ERROR = 4;
var APEX4_SOCKET_UNAUTHORIZED = 5;
var Apex4SocketConnectionStatuses = ["NOT DEFINED", "UNKNOWN", "CONNECTED", "DISCONNECTED", "ERROR", "UNAUTHORIZED"];

var APEX4_HOST_UNDEFINED = 0;
var APEX4_HOST_VALID = 1;
var Apex4HostStatuses = ["UNDEFINED", "VALID"];

var LogManager = null;
var SocketIoManager = null;
var DigitalInputsManager = null;
var DigitalOutputsManager = null;
var AnalogInputsManager = null;
var AnalogOutputsManager = null;

//#endregion Dependencies and Globals

// Initialize
//#region Initialize
exports.InitializeApex4Instance = function (logger, io) {
    // Initialize the globals
    LogManager = logger;
    SocketIoManager = io;

    // Get the Apex 4 Host information 
    ReadApex4ConfigurationFile(function (apex4Info) {
        Apex4HostInfo = apex4Info;
    });
    LogManager.info("Apex4.InitializeApex4Instance: Apex 4 instance started: " + JSON.stringify(Apex4HostInfo));
}

exports.InitializeApex4IoInterfaces = function (digitalInputs, digitalOutputs, analogInputs, analogOutputs) {
    // Initialize the I/O managers
    DigitalInputsManager = digitalInputs;
    DigitalOutputsManager = digitalOutputs;
    AnalogInputsManager = analogInputs;
    AnalogOutputsManager = analogOutputs;

    // If we have a valid token, we're ready to talk to Apex 4
    if (Apex4HostInfo.Valid == APEX4_HOST_VALID) {
        CloseApex4Connection();// Close and destroy any previous connection
        SetApex4Structure(Apex4HostInfo.Token);// Apex 4 requires a structure of how I/O is organized
        ConnectToApex4Host();// Start the Apex 4 connection
    } else {
        // Since we don't have a valid host, mark the connection status as 'NOT DEFINED'
        SetApex4ConnectionStatus(APEX4_SOCKET_NOTDEFINED);
    }
    LogManager.info("Apex4.InitializeApex4Instance: Apex 4 I/O interfaces initialized");
}

function ReadApex4ConfigurationFile(callback) {
    var apex4Info;
    try {
        // Try to read the Apex 4 file ...
        apex4Info = require("./apex4.json");
        LogManager.info("Apex4.ReadApex4ConfigurationFile: Apex 4 settings=" + JSON.stringify(apex4Info));
        callback(apex4Info);
    }
    catch (error) {
        // Something went wrong ... assume that we have not defined a file to this point, or the file is corrupt
        //  in either case, write a new good file that we can use as a template for later, but with Apex 4 host as undefined
        ResetApex4Configuration(function (resetApex4info) {
            LogManager.info("Apex4.ReadApex4ConfigurationFile: Apex 4 settings after reset=" + JSON.stringify(resetApex4info));
            callback(resetApex4info);
        });
    }
}

function ResetApex4Configuration(callback) {
    var resetApex4Info = { "Address": "", "Port": "", "Token": "", "Valid": APEX4_HOST_UNDEFINED };
    fs.writeFile("./apex4.json", JSON.stringify(resetApex4Info), 'utf8', function (fserror) {
        if (fserror) {
            LogManager.error("Apex4.ReadApex4ConfigurationFile: Unable to write new Apex 4 host configuration. " + fserror);
        }
        // Do a sync even if error ...
        exec("sync", function (exerror, stdout, stderr) {
            if (exerror) {
                LogManager.error("Apex4.ReadApex4ConfigurationFile: Unable to sync new Apex 4 host configuration. Error=" + exerror);
            }
            LogManager.info("Apex4.ReadApex4ConfigurationFile: Reset Apex 4 host configuration to empty:" + JSON.stringify(resetApex4Info));
            SetApex4ConnectionStatus(APEX4_SOCKET_NOTDEFINED);
            callback(resetApex4Info);
        });
    });
}

function SetApex4Structure(token) {

    Apex4Structure = { "token": "", "Inputs": [], "Outputs": [] };// Re-initialize to empty
    Apex4Structure.token = token;

    var _digitalInputsArray = DigitalInputsManager.GetDigitalInputsArray();
    for (var io = 0; io < _digitalInputsArray.length; io++) {
        if (_digitalInputsArray[io].Enabled === "TRUE") {
            Apex4Structure.Inputs.push({
                "uid": "DI" + _digitalInputsArray[io].Address,
                "name": _digitalInputsArray[io].Name,
                "type": "bool"
            });
        }
    }
    var _analogInputsArray = AnalogInputsManager.GetAnalogInputsRealtime();
    for (var io = 0; io < _analogInputsArray.length; io++) {
        var units = "V";
        if (_analogInputsArray[io].FormulaId != -1) {
            units=AnalogInputsManager.GetAnalogInputUnits(io);
        }
        if (_analogInputsArray[io].Enabled === "TRUE") {
            Apex4Structure.Inputs.push({
                "uid": "AI" + _analogInputsArray[io].Address,
                "name": _analogInputsArray[io].Name,
                "type": "decimal",
                "units": units
            });
        }
    }
    var _digitalOutputsArray = DigitalOutputsManager.GetDigitalOutputsArray();
    for (var io = 0; io < _digitalOutputsArray.length; io++) {
        if (_digitalOutputsArray[io].Enabled === "TRUE") {
            Apex4Structure.Outputs.push({
                "uid": "DO" + _digitalOutputsArray[io].Address,
                "name": _digitalOutputsArray[io].Name,
                "type": "bool"
            });
        }
    }
    var _analogOutputsArray = AnalogOutputsManager.GetAnalogOutputsArray();
    for (var io = 0; io < _analogOutputsArray.length; io++) {
        if (_analogOutputsArray[io].Enabled === "TRUE") {
            Apex4Structure.Outputs.push({
                "uid": "AO" + _analogOutputsArray[io].Address,
                "name": _analogOutputsArray[io].Name,
                "type": "decimal",
                "units": "V"
            });
        }
    }
    LogManager.info("Apex4.SetApex4Structure: Apex4Structure=" + JSON.stringify(Apex4Structure));
}
//#endregion Initialize

//#region Apex 4 Connection
function ConnectToApex4Host() {
    //#region Update I/O Modules with Apex 4 token and socket
    Apex4SocketConnection = require('socket.io/node_modules/socket.io-client')('https://' + Apex4HostInfo.Address + ':' + Apex4HostInfo.Port + '?token=' + Apex4HostInfo.Token, { 'forceNew': true });
    SetApex4ConnectionStatus(APEX4_SOCKET_UNKNOWN);// Default to unknown

    DigitalInputsManager.ConfigureApex4TokenAndSocket(Apex4HostInfo.Token, Apex4SocketConnection);
    DigitalOutputsManager.ConfigureApex4TokenAndSocket(Apex4HostInfo.Token, Apex4SocketConnection);
    AnalogInputsManager.ConfigureApex4TokenAndSocket(Apex4HostInfo.Token, Apex4SocketConnection);
    AnalogInputsManager.ResetApex4PreviousValues();
    AnalogOutputsManager.ConfigureApex4TokenAndSocket(Apex4HostInfo.Token, Apex4SocketConnection);
    //#endregion Update I/O Modules with Apex 4 token and socket

    //#region Apex 4 Client socket messaging
    // Dude, we're connected ... here's your data
    LogManager.warn('Apex4.SetApex4Connection: Initiating Apex 4 connection to ' + Apex4HostInfo.Address + ':' + Apex4HostInfo.Port);
    Apex4SocketConnection.on('connect', function () {

        Apex4SocketConnection.off('authorized');
        Apex4SocketConnection.on('authorized', function () {

            // Hallelujah, we've connected to the great god Apex 4, our problems are solved!
            SetApex4ConnectionStatus(APEX4_SOCKET_CONNECTED);
            LogManager.info('Apex4.SetApex4Connection: Connected to Apex 4 at ' + Apex4HostInfo.Address + ':' + Apex4HostInfo.Port);

            // Keep alive from server ... just respond
            Apex4SocketConnection.off('check');
            Apex4SocketConnection.on('check', function (message) {
                Apex4SocketConnection.emit('keepalive', message);
            });

            // Send back the data structure
            Apex4SocketConnection.off('requestsync');
            Apex4SocketConnection.on('requestsync', function () {
                SetApex4Structure(Apex4HostInfo.Token);
                LogManager.info("Apex4.SetApex4Connection: Apex 4 sync requested, returning=" + JSON.stringify(Apex4Structure));
                Apex4SocketConnection.emit("sync", Apex4Structure);
            });

            Apex4SocketConnection.on('changestate', function () {
                LogManager.fatal("Apex4.SetApex4Connection: changestate=" + JSON.stringify(message));
            });
        });

        // Token must be wrong or server didn't understand
        Apex4SocketConnection.off('unauthorized');
        Apex4SocketConnection.on('unauthorized', function (msg) {
            LogManager.error("Apex4.SetApex4Connection: Unauthorized access to Apex 4: " + msg);
            SetApex4ConnectionStatus(APEX4_SOCKET_ERROR);// Uh-oh!
        });

        // Dude, we're leaving, we may be back
        Apex4SocketConnection.off('disconnect');
        Apex4SocketConnection.on('disconnect', function (a) {
            LogManager.warn("Apex4.SetApex4Connection: Disconnected from Apex 4");
            SetApex4ConnectionStatus(APEX4_SOCKET_DISCONNECTED);// Oops!
        });
    });
    //#endregion Apex 4 Client socket messaging
}

function SetApex4ConnectionStatus(status) {
    Apex4SocketConnectionStatus = status;
    SocketIoManager.sockets.emit("Apex4Status", { Apex4Status: Apex4SocketConnectionStatuses[status] });
}
//#endregion Apex 4 Connection

//#region API interface
exports.GetApex4HostInfo = function () {
    return Apex4HostInfo;
}

exports.GetApex4HostConnectionStatus = function () {
    return Apex4SocketConnectionStatuses[Apex4SocketConnectionStatus];
}

exports.SetApex4Host = function (apex4Info, callback) {
    // Update the Apex 4 Host Info
    Apex4HostInfo = apex4Info;
    Apex4HostInfo.Valid = APEX4_HOST_VALID;// Assume it's valid because we checked the parameters at entry

    // Write the host information to the file
    fs.writeFile("./apex4.json", JSON.stringify(Apex4HostInfo), 'utf8', function (fserror) {
        if (fserror) {
            LogManager.error("Apex4.Apex4HostConnection: Unable to write Apex 4 host configuration. " + fserror);
        }
        // Do a sync even if error ...
        exec("sync", function (exerror, stdout, stderr) {
            if (exerror) {
                LogManager.error("Apex4.Apex4HostConnection: Unable to sync Apex 4 host configuration. Error=" + exerror);
            }
        });
    });

    // Create the new Apex 4 connection
    CloseApex4Connection();// Close and destroy any previous connection
    SetApex4Structure(Apex4HostInfo.Token);// Apex 4 requires a structure of how I/O is organized
    ConnectToApex4Host();// Events will be generated as the new connection comes up

    // Return the updated Apex 4 host info
    callback(Apex4HostInfo);
}

function CloseApex4Connection() {
    // Check to see if there a connection ... if so, close it.
    if (Apex4SocketConnection) {
        try {
            Apex4SocketConnection.emit('end');
            Apex4SocketConnection.disconnect();
            Apex4SocketConnection.close();
        }
        catch (socketError) {
            LogManager.error("Apex4.CloseApex4Connection: Error closing existing Apex4SocketConnection=" + socketError);
        } finally {
            Apex4SocketConnection = null;
        }
    }
}

exports.RemoveApex4Host = function (callback) {
    CloseApex4Connection();// Close and destroy any previous connection
    ResetApex4Configuration(function (resetApex4Info) {
        callback(resetApex4Info);
    });// Write and then return an empty connection
}
//#endregion API interface
